/*
 *  Copyright (C) 2000-2006  Henk de Groot - PE1DNN
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *  Please note that the GPL allows you to use the driver, NOT the radio.
 *  In order to use the radio, you need a license from the communications
 *  authority of your country.
 */

#include <ctype.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <termio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/io.h>
#include "linux.h"
#include "digicons.h"

/* global data */
short directvideo;

/* local data */
static int    kb_initialized = 0;
static int    kb_fd   = -1;
static int    kb_char = -1;
static int    kb_esc  = 0;
static struct termios t_old;

#ifdef _TELEMETRY_
/*
 * Parameter as the verified format "lpt1", "lpt2" or "lpt3", or
 * for the 8 bit multiplexed input "lpt1_8", "lpt2_8" or "lpt3_8".
 * The function does not check this but assumes it is okay!.
 *
 * The second parameter is used in case the lpt port is multiplexed.
 * The port number (which is 0 based and runs from 0 to 7) is put on
 * bits 1, 2 and 3 of the control port. These bits correspond with the
 * "Auto Linefeed", "Initialize Printer" and "Select Printer" outputs
 * of the LPT port.
 *
 * Function returns the LPT status or -1 if the status could not be
 * read.
 */
short get_lpt_status(char *lpt, short port)
{
    short status;
    short lptio;
    int   arg;

#ifdef for_testing_wx_without_interface
    int   res;

    FILE *fd;
    char  line[80];

    if(lpt[3] == '1') {
        fd = fopen("../data.txt", "r");

        fgets(line, 79, fd);
        while(port > 0)
        {
            fgets(line, 79, fd);
            port--;
        }

        fclose(fd);

        sscanf(line, "%d", &res);

        fflush(stdout);

        return res;
    }
#endif /* for_testing_wx_without_interface */

    /* switch on lpt number */
    /*
     * Port numbers are taken in the same order as in
     *
     * .../linux/drivers/char/lp.c
     *
     * at least in Linux kernel 2.0.36.
     *
     * If you changed the driver in Linux to other ports then you will
     * most likely know how to make changes here too :-)
     */
    switch(lpt[3]) {
    case '1':
            lptio = 0x3bc;
            break;
    case '2':
            lptio = 0x378;
            break;
    case '3':
            lptio = 0x278;
            break;
    default:
            return -1;
    }

#ifndef __CYGWIN32__
    if(ioperm(lptio, 3, 1) < 0)
    {
        /* Can't get permission to access the port */
        return -1;
    }
#endif

    /* shift port number and keep the 3 port-number bits */
    port = (port << 1) & 0x0e;

    /*
     * Invert the "Auto Linefeed" and "Printer Select" bits, these
     * are inverted back by hardware.
     */
    port = port ^ 0x0a;

    /*
     * Make sure the port is not bi-directional (needed for some
     * systems, for example the HP OmniBook 600C)
     */
    /*
     * write control port with the port number and resetting
     * directonal bits and strobe.
     */
    outb(port, lptio + 2);

    /*
     * delay for 4 ms to give the hardware time to put the nibble
     * on the LPT port
     */
    usleep(4000L);

    /*
     * Read status register, one above the base port
     */
    if(lpt[4] == '\0')
    {
        arg = inb(lptio + 1);

        /* move read character to a short */
        status = (short) arg;

        /* one bit was reversed by the port hardware, flip it back */
        status = status ^ 0x080;

        /* make shure the top byte is 0 and the 3 lower bits are also 0 */
        status = status & 0x00f8;
    }
    else
    {
        /* set strobe bit and port number */
        outb(port + 0x01, lptio + 2);

        /*
         * delay for 4 ms to give the hardware time to put the nibble
         * on the LPT port
         */
        usleep(4000L);

        /* read lower nibble */
        arg = (inb(lptio + 1)) &  0x00f0;

        /* make this the lower nibble */
        status = arg >> 4;

        /* reset strobe bit, keep port number */
        outb(port, lptio + 2);

        /*
         * delay for 4 ms to give the hardware time to put the nibble
         * on the LPT port
         */
        usleep(4000L);

        /* read higher nibble */
        arg = (inb(lptio + 1)) &  0x00f0;

        /* add high nibble */
        status = status | arg;

        /*
         * one bit was reversed by the port hardware (in low and high
         * nibble), flip it back
         */
        status = status ^ 0x088;
    }

    /*
     * Give up permissions. We don't care if this fails...
     */
#ifndef __CYGWIN32__
    (void) ioperm(lptio, 3, 0);
#endif

    /* return the status */
    return (status & 0x00ff);
}
#endif /* _TELEMETRY_ */

#ifdef _OUTPORT_
/*
 * Parameter as the verified format "lpt1", "lpt2" or "lpt3", or for the
 * 8 bit multiplexed input "lpt1_8", "lpt2_8" or "lpt3_8". The function
 * does not check this but assumes it is okay!.
 *
 * The data is a string with "0", "1" and pulse and don't care flags, low-bit
 * first. Also the data is already checked before calling this function.
 *
 * Function returns the LPT data or -1 if the data could not be changed.
 */
short out_lpt_data(char *lpt, char *change_data)
{
    char  data[9];
    short lptio;
    short i;
    int   num;
    int   res;
    int   loops;

    /* switch on lpt number */
    /*
     * Port numbers are taken in the same order as in
     *
     * .../linux/drivers/char/lp.c
     *
     * at least in Linux kernel 2.0.36.
     *
     * If you changed the driver in Linux to other ports then you will
     * most likely know how to make changes here too :-)
     */
    switch(lpt[3]) {
    case '1':
            lptio = 0x3bc;
            break;
    case '2':
            lptio = 0x378;
            break;
    case '3':
            lptio = 0x278;
            break;
    default:
            /* Wrong lpt name */
            return -1;
    }

#ifndef __CYGWIN32__
    if(ioperm(lptio, 3, 1) < 0)
    {
        /* Can't get permission to access the port */
        return -1;
    }
#endif

    /*
     * Make sure the port is not bi-directional (needed for some
     * systems, for example the HP OmniBook 600C)
     */
    /* write control port, resetting all bits */
    outb(0, lptio + 2);

    /*
     * delay for 4 ms to give the hardware time to stabilize the data
     * on the LPT port
     */
    usleep(4000L);

    /*
     * Read data register, this is on the base port
     */
    num = inb(lptio);

    /*
     * Convert read data to data string
     */
    for(i = 0; i < 8; i++)
    {
        if((num & 0x01) == 0)
        {
            data[i] = '0';
        }
        else
        {
            data[i] = '1';
        }
        num = num >> 1;
    }
    data[8] = '\0';

    if(strchr(change_data, 'p') == NULL)
    {
        loops = 1;
    }
    else
    {
        loops = 2;
    }

    while(loops > 0)
    {
        /*
         * Take over the not don't care values from change_data into data
         */
        for(i = 0; i < 8; i++)
        {
            if(change_data[i] == '\0')
            {
                /* no more changes, we can stop this loop */
                break;
            }

            if((change_data[i] == '0') || (change_data[i] == '1'))
            {
                /* this bit should be changed */
                data[i] = change_data[i];
            }

            if(change_data[i] == 'p')
            {
                /*
                 * This bit should be flipped (the second time it is flipped
                 * back
                 */
                if(data[i] == '1')
                {
                    data[i] = '0';
                }
                else
                {
                    data[i] = '1';
                }
            }
        }

        /* Convert data to number, argument is D0..D7 */
        num = 0;
        for(i = 7; i >= 0; i--)
        {
            num = num << 1;

            if(data[i] == '1')
            {
                num = num + 1;
            }
        }

        /* Set the data on the LPT port */
        outb(num, lptio);

        /* engage in next loop if needed */
        loops--;
        if(loops > 0)
        {
            /* toggle waits 1 second */
            sleep(1);
        }
    }

    /*
     * delay for 4 ms to give the hardware time to stabilize the data
     * on the LPT port
     */
    usleep(4000L);

    /* Read the data back from the LPT port */
    res = inb(lptio);

    /*
     * Give up permissions. We don't care if this fails...
     */
#ifndef __CYGWIN32__
    (void) ioperm(lptio, 3, 0);
#endif

    /* Check if the data was succesfully changed */
    if(res == num)
    {
        /* return port value for success */
        return ((short) res) & 0x00ff;
    }
    else
    {
        return -1;
    }
}
#endif /* _OUTPORT_ */

/*
 * Restore the tty to the old state we had before initialization
 */
void kb_exit()
{
    if(kb_initialized == 1)
    {
        (void) tcsetattr(kb_fd, TCSANOW, &t_old);
        close(kb_fd);
        kb_initialized = 0;
    }
}

/*
 * Initialize the tty to poll for keys
 *
 * If is fails just silently ignore. The keys will not work but the DIGI,
 * which is most important, will still run. Failures will occur if
 * stdin is not associated with a tty.
 */
void kb_init() 
{
    struct termios t;

    if(kb_initialized == 1)
    {
        /* Already done, ignore */
        return;
    }

    kb_fd = open("/dev/tty", O_RDWR|O_NOCTTY|O_NONBLOCK);

    /* Open our TTY */
    if (kb_fd == -1)
    {
        /* Unable to open, ignore */
        return;
    }

    /* Check if stdin is connected to a tty, stdin has file descriptor 0 */
    if(isatty(kb_fd) == 0)
    {
        /* Not a TTY, ignore */
        close(kb_fd);
        return;
    }

    /* Get the attributes of the TTY for modification */
    if (tcgetattr(kb_fd, &t) != 0)
    {
        /* Can't get settings, kb not initialized */
        close(kb_fd);
        return;
    }

    /* Get the attributes of the TTY to be able to restore them later */
    if (tcgetattr(kb_fd, &t_old) != 0)
    {
        /* Can't get settings, kb not initialized */
        close(kb_fd);
        return;
    }

    /* Set the terminal settings to raw access */
    t.c_cc[VMIN]  = 0;
    t.c_cc[VTIME] = 2;
    t.c_iflag     = 0;
    t.c_lflag     = ISIG;
    t.c_cflag    |= HUPCL|CLOCAL|CREAD|CRTSCTS;

    /* flush the old data */
    if (tcflush(kb_fd, TCIFLUSH) == -1)
    {
        /* Can't flush old data, kb not initialized */
        close(kb_fd);
        return;
    }

    /* Set the new attributes for the TTY */
    if (tcsetattr(kb_fd, TCSANOW, &t) == -1)
    {
        /* Can't change settings, kb not initialized */
        close(kb_fd);
        return;
    }

    /* Intialize used variables */
    kb_char = -1;
    kb_esc  = 0;

    /* Change flag, we are done! */
    kb_initialized = 1;

    return;
}

/*
 * Check if the keyboard has been hit
 */
int kbhit()
{
    int    i;
    char   key;

    if(kb_initialized != 1)
    {
        /*
         * We don't read keys, initialization failed somehow or
         * initialization was not done because the keyboard support
         * was not wanted.
         */
        return 0;
    }

    i = read(kb_fd, &key, 1);

    if(i == -1)
    {
        if(kb_char != -1)
        {
            /* still key pending (after ESC) */
            return 1;
        }
        return 0;
    }
    else
    {
        /* Use ESC as "ALT" replacement */
        if(key == 0x001b)
        {
            if(kb_esc == 0)
            {
                /* First escape, remember this! No key yet */
                kb_esc = 1;
                return 0;
            }
            else
            {
                /* Second escape, we want this key! */
                kb_esc  = 0;
                kb_char = key;
                return 1;
            }
        }
        else 
        {
            /*
             * Key was not ESC. Was previous key ESC? if if no then just
             * pass, if yes, translate to DOS scancode
             */
            if(kb_esc == 0)
            {
#ifdef __CYGWIN32__
                /* ALT key conversions */
                switch((int) key) {
                case -31: /* ALT A */
                          kb_esc = 1;
                          kb_char = 0x1e;
                          return kb_esc;
                          break;
                case -30: /* ALT B */
                          kb_esc = 1;
                          kb_char = 0x30;
                          return kb_esc;
                          break;
                case -24: /* ALT H */
                          kb_esc = 1;
                          kb_char = 0x23;
                          return kb_esc;
                          break;
                case -20: /* ALT L */
                          kb_esc = 1;
                          kb_char = 0x26;
                          return kb_esc;
                          break;
                case -12: /* ALT T */
                          kb_esc = 1;
                          kb_char = 0x14;
                          return kb_esc;
                          break;
                case -10: /* ALT V */
                          kb_esc = 1;
                          kb_char = 0x2f;
                          return kb_esc;
                          break;
                case -8:  /* ALT X */
                          kb_esc = 1;
                          kb_char = 0x2d;
                          return kb_esc;
                          break;
                }
#endif
                /* not following ESC, normal key */

                /* Map 0x7f to backspace */
                if(key == 0x7f)
                        key = '\b';

                kb_char = key;
                return 1;
            }
            else
            {
                /* following ESC, ALT key, convert */
                switch(key) {
                case 'a':
                case 'A':   kb_char = 0x1e;
                            break;
                case 'b':
                case 'B':   kb_char = 0x30;
                            break;
                case 'h':
                case 'H':   kb_char = 0x23;
                            break;
                case 'l':
                case 'L':   kb_char = 0x26;
                            break;
                case 't':
                case 'T':   kb_char = 0x14;
                            break;
                case 'v':
                case 'V':   kb_char = 0x2f;
                            break;
                case 'x':
                case 'X':   kb_char = 0x2d;
                            break;
                default:    /* remove ESC flag */
                            kb_esc = 0;
                }

                /* return that we have a complete sequence now */
                return kb_esc;
            }
        }
    }
}

/*
 * Return a character after kbhit returned '1'
 */
char getch()
{
    char c;

    if(kb_esc == 1)
    {
        kb_esc = 0;
        return '\0';
    }
    else
    {
        c = (char) kb_char; 
        kb_char = -1;

        return c;
    }
}

#ifdef _SERIAL_
/*
 * Restore the tty to the old state we had before initialization
 */
void serial_close(int fd)
{
    struct termios t;

    if(fd != -1)
    {
        /* Get the attributes of the serial line for modification */
        if (tcgetattr(fd, &t) != 0)
        {
            /* Can't get settings, close and quit without restoring */
            close(fd);
            return;
        }

        /* Set the terminal back in canonical mode with signal support*/
        t.c_lflag     = ISIG | ICANON;

        /* flush the old data */
        (void) tcflush(fd, TCIFLUSH);

        /* Set the new attributes for the serial line */
        (void) tcsetattr(fd, TCSANOW, &t);

        close(fd);
    }
}

/*
 * Open the serial line to poll for data
 *
 * Return -1 if it fails.
 */
int serial_open(short comnr, short baudrate) 
{
    struct termios t;
    char   dev[256];
    int    fd;

    sprintf(dev, "/dev/ttyS%d", comnr - 1);

    fd = open(dev, O_RDWR|O_NOCTTY|O_NONBLOCK);

    /* Open our serial line */
    if (fd == -1)
    {
        /* Unable to open, ignore */
        return -1;
    }

    /* Check if fd is connected to a tty, stdin has file descriptor 0 */
    if(isatty(fd) == 0)
    {
        /* Not a serial line, ignore */
        close(fd);
        return -1;
    }

    /* Get the attributes of the serial line for modification */
    if (tcgetattr(fd, &t) != 0)
    {
        /* Can't get settings, not initialized */
        close(fd);
        return -1;
    }

    /* Set the terminal settings to raw access */
    t.c_cc[VMIN]  = 0;
    t.c_cc[VTIME] = 2;
    t.c_iflag     = 0;
    t.c_lflag     = 0;
    t.c_cflag     = CS8 | HUPCL|CLOCAL|CREAD;

    /* Change the baudrate settings */
    switch(baudrate) {
    case 1200: cfsetispeed(&t, B1200);
               cfsetospeed(&t, B1200);
               break;
    case 2400: cfsetispeed(&t, B2400);
               cfsetospeed(&t, B2400);
               break;
    case 4800: cfsetispeed(&t, B4800);
               cfsetospeed(&t, B4800);
               break;
    case 9600: cfsetispeed(&t, B9600);
               cfsetospeed(&t, B9600);
               break;
    default:   /* Illegal baudrate on serial, ignore */
               break;
    }

    /* flush the old data */
    if (tcflush(fd, TCIFLUSH) == -1)
    {
        /* Can't flush old data, serial line not initialized */
        serial_close(fd);
        return -1;
    }

    /* Set the new attributes for the serial line */
    if (tcsetattr(fd, TCSANOW, &t) == -1)
    {
        /* Can't change settings, serial line not initialized */
        serial_close(fd);
        return -1;
    }

    return fd;
}

/*
 * Read a character from the serial line
 */
int serial_in(int fd, char *buf_p)
{
    int i;

    i = read(fd, buf_p, 1);

    return i;
}
#endif /* _SERIAL_ */
